from mongoframes.factory import quotas
from mongoframes.factory.makers import numbers as number_makers

from tests.fixtures import *


def test_counter():
    """`Counter` makers should return a number sequence"""

    # Configured with defaults
    maker = number_makers.Counter()

    for i in range(1, 100):
        # Check the assembled result
        assembled = maker._assemble()
        assert assembled == i

        # Check the finished result
        finished = maker._finish(assembled)
        assert finished == i

    # Configuered with custom start from and step
    maker = number_makers.Counter(quotas.Quota(10), quotas.Quota(5))

    for i in range(10, 100, 5):
        # Check the assembled result
        assembled = maker._assemble()
        assert assembled == i

        # Check the finished result
        finished = maker._finish(assembled)
        assert finished == i

    # Reset should reset the counter to the initial start from value
    maker.reset()

    # Check the assembled result
    assembled = maker._assemble()
    assert assembled == 10

    # Check the finished result
    finished = maker._finish(assembled)
    assert finished == 10

def test_float():
    "`Float` makers should return a float between two values"

    min_value = 2.5
    max_value = 7.2
    maker = number_makers.Float(min_value, max_value)

    for i in range(1, 100):
        # Check the assembled result
        assembled = maker._assemble()
        assert assembled >= min_value and assembled <= max_value

        # Check the finished result
        finished = maker._finish(assembled)
        assert finished >= min_value and assembled <= max_value

def test_int():
    "`Int` makers should return an integer between two values"

    min_value = 25
    max_value = 72
    maker = number_makers.Int(min_value, max_value)

    for i in range(1, 100):
        # Check the assembled result
        assembled = maker._assemble()
        assert assembled >= min_value and assembled <= max_value

        # Check the finished result
        finished = maker._finish(assembled)
        assert finished >= min_value and assembled <= max_value


import random

from mongoframes.factory.makers import Maker

__all__ = [
    'Counter'
    ]


class Counter(Maker):
    """
    Generate a sequence of numbers.
    """

    def __init__(self, start_from=1, step=1):
        super().__init__()

        self._start_from = int(start_from)
        self._step = step
        self._counter = self._start_from

    def reset(self):
        self._counter = int(self._start_from)

    def _assemble(self):
        value = self._counter
        self._counter += int(self._step)
        return value


class Float(Maker):
    """
    Generate a random float between two values.
    """

    def __init__(self, min_value, max_value):
        super().__init__()

        self._min_value = min_value
        self._max_value = max_value

    def _assemble(self):
        return random.uniform(float(self._min_value), float(self._max_value))


class Int(Float):
    """
    Generate a random integer between two values.
    """

    def _assemble(self):
        return random.randint(int(self._min_value), int(self._max_value))


from blinker import signal
from collections import OrderedDict

from mongoframes.factory.makers import Maker

__all__ = ['Blueprint']


class _BlueprintMeta(type):
    """
    Meta class for `Frame`s to ensure an `_id` is present in any defined set of
    fields.
    """

    def __new__(meta, name, bases, dct):

        # Collect all instructions for the blueprint. Each instruction is stored
        # as a key value pair in a dictionary where the name represents the
        # field the instruction refers to and the value is a `Maker` type for
        # generating a value for that field.
        dct['_instructions'] = dct.get('_instructions') or {}

        # Ensure the instructions are an ordered dictionary
        dct['_instructions'] = OrderedDict(dct['_instructions'])

        for k, v in dct.items():
            if isinstance(v, Maker):
                dct['_instructions'][k] = v

        # Check the blueprint has a frame class associated with it. The `Frame`
        # class will be used to insert the document into the database.
        assert '_frame_cls' in dct or len(bases) == 0, \
                'No `_frame_cls` defined for the blueprint'

        # Check for a set of meta fields or define an empty set if there isn't
        # one. Meta-fields determine a set of one or more fields that should be
        # set when generating a document but which are not defined in the
        # assoicated `Frame` classes `_fields` attribute.
        if '_meta_fields' not in dct:
            dct['_meta_fields'] = set([])

        return super(_BlueprintMeta, meta).__new__(meta, name, bases, dct)


class Blueprint(metaclass=_BlueprintMeta):
    """
    Blueprints provide the instructions for producing a fake document for a
    collection represented by a `Frame` class.
    """

    def __init__(self):
        assert False, \
            'Blueprint classes should remain static and not be initialized'

    # Public methods

    @classmethod
    def get_frame_cls(cls):
        """Return the `Frame` class for the blueprint"""
        return cls._frame_cls

    @classmethod
    def get_instructions(cls):
        """Return the instructions for the blueprint"""
        return dict(cls._instructions)

    @classmethod
    def get_meta_fields(cls):
        """Return the meta-fields for the blueprint"""
        return cls._meta_fields

    # Factory methods

    @classmethod
    def assemble(cls):
        """Assemble a single document using the blueprint"""
        document = {}
        for field_name, maker in cls._instructions.items():
            with maker.target(document):
                document[field_name] = maker()
        return document

    @classmethod
    def finish(cls, document):
        """
        Take a assembled document and convert all assembled values to
        finished values.
        """
        document_copy = {}
        for field_name, value in document.items():
            maker = cls._instructions[field_name]
            with maker.target(document):
                document_copy[field_name] = maker(value)
        return document_copy

    @classmethod
    def reassemble(cls, fields, document):
        """
        Take a previously assembled document and reassemble the given set of
        fields for it in place.
        """
        for field_name in cls._instructions:
            if field_name in fields:
                maker = cls._instructions[field_name]
                with maker.target(document):
                    document[field_name] = maker()

    @classmethod
    def reset(cls):
        """
        Reset the blueprint. Blueprints are typically reset before being used to
        assemble a quota of documents. Resetting a Blueprint will in turn reset
        all the Maker instances defined as instructions for the Blueprint
        allowing internal counters and alike to be reset.
        """
        # Reset instructions
        for maker in cls._instructions.values():
            maker.reset()

    # Events

    @classmethod
    def on_fake(cls, frames):
        """Called before frames are inserted"""

        # By default the hook will simply trigger a `fake` event against the
        # frame. This allows the overriding method to control this behaviour.
        signal('fake').send(cls._frame_cls, frames=frames)

    @classmethod
    def on_faked(cls, frames):
        """Called after frames are inserted"""

        # By default the hook will simply trigger a `fake` event against the
        # frame. This allows the overriding method to control this behaviour.
        signal('faked').send(cls._frame_cls, frames=frames)


"""
Support for paginating frames.
"""

from copy import deepcopy
import math

from mongoframes.queries import Condition, Group, to_refs


__all__ = (
    # Exceptions
    'InvalidPage',

    # Classes
    'Page',
    'Paginator'
    )


class InvalidPage(Exception):
    """
    An error raised when an invalid page is requested.
    """


class Page(object):
    """
    A class to represent one page of results.
    """

    def __init__(self, offset, number, items, next, prev):

        # The offset of the first result in the page from the first result of
        # the entire selection.
        self._offset = offset

        # The page number
        self._number = number

        # The results/frames for this page
        self._items = list(items)

        # The next and previous page numbers (if there is no next/previous page
        # the value will be None.
        self._next = next
        self._prev = prev

    def __getitem__(self, i):
        return self._items[i]

    def __iter__(self):
        for item in self._items:
            yield item

    def __len__(self):
        return len(self._items)

    # Read-only properties

    @property
    def items(self):
        """Return a list of results for the page"""
        return self._items

    @property
    def next(self):
        """
        Return the page number for the next page or None if there isn't one.
        """
        return self._next

    @property
    def number(self):
        """Return the page number"""
        return self._number

    @property
    def prev(self):
        """
        Return the page number for the previous page or None if there isn't one.
        """
        return self._prev

    # Public methods

    def offset(self, item):
        """Return the offset for an item in the page"""
        return self._offset + self.items.index(item)


class Paginator(object):
    """
    A pagination class for slicing query results into pages. This class is
    designed to work with Frame classes.
    """

    def __init__(
            self,
            frame_cls,
            filter=None,
            per_page=20,
            orphans=0,
            **filter_args
            ):

        # The frame class results are being paginated for
        self._frame_cls = frame_cls

        # The filter applied when selecting results from the database (we
        # flattern the filter at this point which effectively deep copies.
        if isinstance(filter, (Condition, Group)):
            self._filter = filter.to_dict()
        else:
            self._filter = to_refs(filter)

        # Any additional filter arguments applied when selecting results such as
        # sort and projection,
        self._filter_args = filter_args

        # The number of results that will be displayed per page
        self._per_page = per_page

        # If a value is specified for orphans then the last page will be able to
        # hold the additional results (up to the value of orphans). This can
        # help prevent users being presented with pages contain only a few
        # results.
        self._orphans = orphans

        # Count the total results being paginated
        self._items_count = frame_cls.count(self._filter)

        # Calculated the number of pages
        total = self._items_count - orphans
        self._page_count = max(1, int(math.ceil(total / float(self._per_page))))

        # Create a list of page number that can be used to navigate the results
        self._page_numbers = range(1, self._page_count + 1)

    def __getitem__(self, page_number):
        if page_number not in self._page_numbers:
            raise InvalidPage(page_number, self.page_count)

        # Calculate the next and previous page numbers
        next = page_number + 1 if page_number + 1 in self._page_numbers else None
        prev = page_number - 1 if page_number - 1 in self._page_numbers else None

        # Select the items for the page
        filter_args = deepcopy(self._filter_args) or {}
        filter_args['skip'] = (page_number - 1) * self._per_page
        filter_args['limit'] = self._per_page

        # Check to see if we need to account for orphans
        if self.item_count - (page_number * self._per_page) <= self.orphans:
            filter_args['limit'] += self.orphans

        # Select the results for the page
        items = self._frame_cls.many(self._filter, **filter_args)

        # Build the page
        return Page(
            offset=filter_args['skip'],
            number=page_number,
            items=items,
            next=next,
            prev=prev
            )

    def __iter__(self):
        for page_number in self._page_numbers:
            yield self[page_number]

    # Read-only properties

    @property
    def item_count(self):
        """Return the total number of items being paginated"""
        return self._items_count

    @property
    def orphans(self):
        """
        Return the number of orphan results that will be allowed in the last
        page of results.
        """
        return self._orphans

    @property
    def page_count(self):
        """Return the total number of pages"""
        return self._page_count

    @property
    def page_numbers(self):
        """Return a list of page numbers"""
        return self._page_numbers

    @property
    def per_page(self):
        """
        Return the number of results per page (with the exception of orphans).
        """
        return self._per_page


from mongoframes.factory import makers
from mongoframes.factory import quotas
from mongoframes.factory.makers import text as text_makers


def test_code():
    """`Code` makers should return a random code"""

    # Configured with the default character set
    maker = text_makers.Code(quotas.Quota(4))

    # Check the assembled result
    assembled = maker._assemble()
    assert len(assembled) == 4
    assert set(assembled).issubset(set(maker.default_charset))

    # Check the finished result
    finished = maker._finish(assembled)
    assert len(finished) == 4
    assert set(assembled).issubset(set(maker.default_charset))

    # Configured with a custom charset
    custom_charset = 'ABCDEF1234567890'
    maker = text_makers.Code(quotas.Quota(6), custom_charset)

    # Check the assembled result
    assembled = maker._assemble()
    assert len(assembled) == 6
    assert set(assembled).issubset(set(custom_charset))

    # Check the finished result
    finished = maker._finish(assembled)
    assert len(finished) == 6
    assert set(assembled).issubset(set(custom_charset))

def test_join():
    """
    `Join` makers should return a the value of one or more items (python strings
    or makers) joined together.
    """

    # Configured with a list of python strings
    maker = text_makers.Join(['foo', 'bar', 'zee'])

    # Check the assembled result
    assembled = maker._assemble()
    assert assembled == ['foo', 'bar', 'zee']

    # Check the finished result
    finished = maker._finish(assembled)
    assert finished == 'foo bar zee'

    # Configured with a list of makers
    maker = text_makers.Join([
        makers.Static('foo'),
        makers.Static('bar'),
        makers.Static('zee')
        ])

    # Check the assembled result
    assembled = maker._assemble()
    assert assembled == ['foo', 'bar', 'zee']

    # Check the finished result
    finished = maker._finish(assembled)
    assert finished == 'foo bar zee'

    # Configured with a custom separator
    maker = text_makers.Join(['foo', 'bar', 'zee'], '-')

    # Check the assembled result
    assembled = maker._assemble()
    assert assembled == ['foo', 'bar', 'zee']

    # Check the finished result
    finished = maker._finish(assembled)
    assert finished == 'foo-bar-zee'

def test_lorem():
    """`Lorem` makers should return lorem ipsum"""

    # Configured to return a body of text
    maker = text_makers.Lorem('body', quotas.Quota(5))

    # Check the assembled result
    assembled = maker._assemble()
    assert len(assembled.split('\n')) == 5

    # Check the finished result
    finished = maker._finish(assembled)
    assert len(finished.split('\n')) == 5

    # Configured to return a paragraph
    maker = text_makers.Lorem('paragraph', quotas.Quota(5))

    # Check the assembled result
    assembled = maker._assemble()
    paragraphs = [p for p in assembled.split('.') if p.strip()]
    assert len(paragraphs) == 5

    # Check the finished result
    finished = maker._finish(assembled)
    paragraphs = [p for p in finished.split('.') if p.strip()]
    assert len(paragraphs) == 5

    # Configured to return a sentence
    maker = text_makers.Lorem('sentence', quotas.Quota(5))

    # Check the assembled result
    assembled = maker._assemble()
    words = [w for w in assembled.split(' ') if w.strip()]
    assert len(words) == 5

    # Check the finished result
    finished = maker._finish(assembled)
    words = [w for w in finished.split(' ') if w.strip()]
    assert len(words) == 5

def test_markov():
    """`Markov` makers should return random text built from a text body"""

    # Set up markov word database
    with open('tests/factory/data/markov.txt') as f:
        text_makers.Markov.init_word_db('test', f.read())

    # Configured to return a body of text
    maker = text_makers.Markov('test', 'body', quotas.Quota(5))

    # Check the assembled result
    assembled = maker._assemble()
    assert len(assembled.split('\n')) == 5

    # Check the finished result
    finished = maker._finish(assembled)
    assert len(finished.split('\n')) == 5

    # Configured to return a paragraph
    maker = text_makers.Markov('test', 'paragraph', quotas.Quota(5))

    # Check the assembled result
    assembled = maker._assemble()
    paragraphs = [p for p in assembled.split('.') if p.strip()]
    assert len(paragraphs) == 5

    # Check the finished result
    finished = maker._finish(assembled)
    paragraphs = [p for p in finished.split('.') if p.strip()]
    assert len(paragraphs) == 5

    # Configured to return a sentence
    maker = text_makers.Markov('test', 'sentence', quotas.Quota(5))

    # Check the assembled result
    assembled = maker._assemble()
    words = [w for w in assembled.split(' ') if w.strip()]
    assert len(words) == 5

    # Check the finished result
    finished = maker._finish(assembled)
    words = [w for w in finished.split(' ') if w.strip()]
    assert len(words) == 5

def test_sequence():
    """
    `Sequence` makers should generate a sequence of strings containing a
    sequence of numbers.
    """

    # Configured with the defaut start from
    maker = text_makers.Sequence('foo-{index}')

    for i in range(0, 10):

        # Check the assembled result
        assembled = maker._assemble()
        assert assembled == 'foo-{index}'.format(index=i + 1)

        # Check the finished result
        finished = maker._finish(assembled)
        assert finished == 'foo-{index}'.format(index=i + 1)

    # Configured with a custom start from
    maker = text_makers.Sequence('foo-{index}', 5)

    for i in range(0, 10):

        # Check the assembled result
        assembled = maker._assemble()
        assert assembled == 'foo-{index}'.format(index=i + 5)

        # Check the finished result
        finished = maker._finish(assembled)
        assert finished == 'foo-{index}'.format(index=i + 5)

    # Reset should reset the sequence to start from
    maker.reset()

    # Check the assembled result
    assembled = maker._assemble()
    assert assembled == 'foo-5'

    # Check the finished result
    finished = maker._finish(assembled)
    assert finished == 'foo-5'


import random
import re
import string

from mongoframes.factory.makers import Faker, Maker

__all__ = [
    'Code',
    'Join',
    'Lorem',
    'Markov'
    'Sequence'
    ]


class Code(Maker):
    """
    Generate a random code of the given length using either the default_charset
    or optionally a custom charset.
    """

    # The default character set that codes are made from
    default_charset = string.ascii_uppercase + string.digits

    def __init__(self, length, charset=None):
        super().__init__()

        # If no charset is provided use the default
        if not charset:
            charset = self.default_charset

        # The length of the code that will be generated
        self._length = length

        # The character set the code will be made from
        self._charset = list(charset)

    def _assemble(self):
        return ''.join([
            random.choice(self._charset) for i in range(0, int(self._length))
            ])


class Join(Maker):
    """
    Join the output of 2 or more items (makers and/or values) together with a
    seperator string.
    """

    def __init__(self, items, sep=' '):
        super().__init__()

        # The list of makers/values that will be joined
        self._items = items

        # The string used to join the items together
        self._sep = sep

    def _assemble(self):
        values = []
        for item in self._items:
            if isinstance(item, Maker):
                values.append(item._assemble())
            else:
                values.append(item)
        return values

    def _finish(self, value):
        parts = []
        for i, item in enumerate(self._items):
            if isinstance(item, Maker):
                parts.append(str(item._finish(value[i])))
            else:
                parts.append(str(item))
        return self._sep.join(parts)


class Lorem(Maker):
    """
    Generate random amounts of lorem ipsum.

    To determine the amount of text generated the type of text structure to
    generate must be specified;

    - body,
    - paragraph,
    - sentence

    along with the quantity;

    - paragraphs in a body,
    - sentences in a paragraph,
    - words in a scentence.
    """

    def __init__(self, text_type, quantity):
        super().__init__()

        # The type of text structure to generate
        self._text_type = text_type

        # The quantity of text to generate
        self._quantity = quantity

        assert self._text_type in ['body', 'paragraph', 'sentence'], \
                'Not a supported text type'

    def _assemble(self):
        quantity = int(self._quantity)

        if self._text_type == 'body':
            return '\n'.join(Faker.get_fake().paragraphs(nb=quantity))

        if self._text_type == 'paragraph':
            return Faker.get_fake().paragraph(
                nb_sentences=quantity,
                variable_nb_sentences=False
                )

        if self._text_type == 'sentence':
            return Faker.get_fake().sentence(
                nb_words=quantity,
                variable_nb_words=False
                )


class Markov(Maker):
    """
    Generate random amounts of text using a Markov chain.

    To determine the amount of text generated the type of text structure to
    generate must be specified;

    - body,
    - paragraph,
    - sentence

    along with the quantity;

    - paragraphs in a body,
    - sentences in a paragraph,
    - words in a sentence.

    This code is heavily based on (lifted from) the code presented in this
    article by Shabda Raaj:
    http://agiliq.com/blog/2009/06/generating-pseudo-random-text-with-markov-chains-u/
    """

    # FIXME:
    #   w1, w2 = w2, random.choice(db['freqs'][(w1, w2)])
    #   KeyError: ('summer...."\n\nTHE', 'END')

    _dbs = {}

    def __init__(self, db, text_type, quantity):
        super().__init__()

        # The database to generate the text from
        self._db = db

        assert db in self.__class__._dbs, 'Word database does not exist'

        # The type of text structure to generate
        self._text_type = text_type

        # The quantity of text to generate
        self._quantity = quantity

        assert self._text_type in ['body', 'paragraph', 'sentence'], \
                'Not a supported text type'

    # Public methds

    @property
    def database(self):
        """Return the selected word database"""
        return self.__class__._dbs[self._db]

    # Private methods

    def _assemble(self):
        quantity = int(self._quantity)

        if self._text_type == 'body':
            return self._body(quantity)

        if self._text_type == 'paragraph':
            return self._paragraph(quantity)

        if self._text_type == 'sentence':
            return self._sentence(quantity)

    def _body(self, paragraphs):
        """Generate a body of text"""
        body = []
        for i in range(paragraphs):
            paragraph = self._paragraph(random.randint(1, 10))
            body.append(paragraph)

        return '\n'.join(body)

    def _paragraph(self, sentences):
        """Generate a paragraph"""
        paragraph = []
        for i in range(sentences):
            sentence = self._sentence(random.randint(5, 16))
            paragraph.append(sentence)

        return ' '.join(paragraph)

    def _sentence(self, words):
        """Generate a sentence"""
        db = self.database

        # Generate 2 words to start a sentence with
        seed = random.randint(0, db['word_count'] - 3)
        seed_word, next_word = db['words'][seed], db['words'][seed + 1]
        w1, w2 = seed_word, next_word

        # Generate the complete sentence
        sentence = []
        for i in range(0, words - 1):
            sentence.append(w1)
            w1, w2 = w2, random.choice(db['freqs'][(w1, w2)])
        sentence.append(w2)

        # Make the sentence respectable
        sentence = ' '.join(sentence)

        # Capitalize the sentence
        sentence = sentence.capitalize()

        # Remove additional sentence ending puntuation
        sentence = sentence.replace('.', '')
        sentence = sentence.replace('!', '')
        sentence = sentence.replace('?', '')
        sentence = sentence.replace(':', '')

        # Remove quote tags
        sentence = sentence.replace('.', '')
        sentence = sentence.replace('!', '')
        sentence = sentence.replace('?', '')
        sentence = sentence.replace(':', '')
        sentence = sentence.replace('"', '')

        # If the last character is not an alphanumeric remove it
        sentence = re.sub('[^a-zA-Z0-9]$', '', sentence)

        # Remove excess space
        sentence = re.sub('\s+', ' ', sentence)

        # Add a full stop
        sentence += '.'

        return sentence

    @classmethod
    def init_word_db(cls, name, text):
        """Initialize a database of words for the maker with the given name"""
        # Prep the words
        text = text.replace('\n', ' ').replace('\r', ' ')
        words = [w.strip() for w in text.split(' ') if w.strip()]

        assert len(words) > 2, \
                'Database text sources must contain 3 or more words.'

        # Build the database
        freqs = {}
        for i in range(len(words) - 2):

            # Create a triplet from the current word
            w1 = words[i]
            w2 = words[i + 1]
            w3 = words[i + 2]

            # Add the triplet to the database
            key = (w1, w2)
            if key in freqs:
                freqs[key].append(w3)
            else:
                freqs[key] = [w3]

        # Store the database so it can be used
        cls._dbs[name] = {
            'freqs': freqs,
            'words': words,
            'word_count': len(words) - 2
            }


class Sequence(Maker):
    """
    Generate a sequence of values where a number is inserted into a template.
    The template should specify an index value, for example:

        "prefix-{index}"
    """

    def __init__(self, template, start_from=1):
        super().__init__()

        self._template = template
        self._start_from = start_from
        self._index = start_from

    def reset(self):
        self._index = self._start_from

    def _assemble(self):
        value = self._template.format(index=self._index)
        self._index += 1
        return value


from functools import reduce
import itertools
import math
import random

from mongoframes.factory.makers import Maker
from mongoframes import ASC

__all__ = [
    'Cycle',
    'OneOf',
    'RandomReference',
    'SomeOf'
    ]


class Cycle(Maker):
    """
    Pick the next item from a list of makers and/or values cycling through the
    list and repeating when we reach the end.
    """

    def __init__(self, items):
        super().__init__()

        # The list of makers/values to select from
        self._items = items

        # The index of the item that will be returned next
        self._item_index = 0

    def reset(self):
        """Reset the item index"""
        self._item_index = 0

    def _assemble(self):
        # Select the next item
        item_index = self._item_index
        item = self._items[item_index]

        # Move the index on 1 (and wrap it if we are at the end of the list)
        self._item_index += 1
        if self._item_index >= len(self._items):
            self._item_index = 0

        # Return the index and it's assembled value
        if isinstance(item, Maker):
            return [item_index, item._assemble()]
        return [item_index, None]

    def _finish(self, value):
        item = self._items[value[0]]
        if isinstance(item, Maker):
            return item._finish(value[1])
        return item


class OneOf(Maker):
    """
    Pick one item from a list of makers and/or values.
    """

    def __init__(self, items, weights=None):
        super().__init__()

        # The list of makers/values to select from
        self._items = items

        # The weighting to apply when selecting a maker/value
        self._weights = weights

    def _assemble(self):
        # Select an item
        item_index = 0
        if self._weights:
            item_index = self.weighted(self._weights)
        else:
            item_index = random.randint(0, len(self._items) - 1)

        # Return the index and it's assembled value
        item = self._items[item_index]
        if isinstance(item, Maker):
            return [item_index, item._assemble()]
        return [item_index, None]

    def _finish(self, value):
        item = self._items[value[0]]
        if isinstance(item, Maker):
            return item._finish(value[1])
        return item

    @staticmethod
    def weighted(weights):
        """
        Return a random integer 0 <= N <= len(weights) - 1, where the weights
        determine the probability of each possible integer.

        Based on this StackOverflow post by Ned Batchelder:

        http://stackoverflow.com/questions/3679694/a-weighted-version-of-random-choice
        """

        # Convert weights to floats
        weights = [float(w) for w in weights]

        # Pick a value at random
        choice = random.uniform(0, sum(weights))

        # Find the value
        position = 0
        for i, weight in enumerate(weights):
            if position + weight >= choice:
                return i
            position += weight


class RandomReference(Maker):
    """
    Pick a reference document at random from a collection (determined by the
    given frame_cls) optionally applying a constraint.
    """

    def __init__(self, frame_cls, constraint=None):
        super().__init__()

        # The frame class that will be used to select a reference with
        self._frame_cls = frame_cls

        # The constraint applied when select a reference document
        self._constraint = constraint or {}

    def _assemble(self):
        # Select a random float that will be used to select a reference document
        # by it's position.
        return random.random()

    def _finish(self, value):
        # Count the number of documents available to pick from
        total_documents = self._frame_cls.count(self._constraint)

        # Calculate the position of the document we've picked
        position = math.floor(total_documents * value)

        # Select the document
        document = self._frame_cls.one(
            self._constraint,
            limit=1,
            skip=position,
            sort=[('_id', ASC)],
            projection={'_id': True}
            )

        # Check the document was found
        if document:
            return document._id

        return None


class SomeOf(Maker):
    """
    Pick one or more items from a list of makers and/or values.
    """

    def __init__(
            self,
            items,
            sample_size,
            weights=None,
            with_replacement=False
            ):
        super().__init__()

        # The list of makers/values to select from
        self._items = items

        # The number of items to pick
        self._sample_size = sample_size

        # The weighting to apply when selecting a maker/value
        self._weights = weights

        # A flag indicating if the same item can be selected from the list more
        # than once.
        self._with_replacement = with_replacement

    def _assemble(self):
        # Select some items
        sample_size = int(self._sample_size)

        sample_indexes = []
        if self._weights:
            sample_indexes = self.weighted(
                self._weights,
                sample_size,
                with_replacement=self._with_replacement
                )
        else:
            sample_range = range(0, len(self._items))
            if self._with_replacement:
                sample_indexes = [random.choice(sample_range) \
                    for s in range(0, sample_size)]
            else:
                sample_indexes = random.sample(sample_range, sample_size)

        # Return the sampled indexes and their assembled values
        values = []
        for sample_index in sample_indexes:
            item = self._items[sample_index]
            if isinstance(item, Maker):
                values.append([sample_index, item._assemble()])
            else:
                values.append([sample_index, None])

        return values

    def _finish(self, value):
        values = []
        for sample in value:
            item = self._items[sample[0]]
            if isinstance(item, Maker):
                values.append(item._finish(sample[1]))
            else:
                values.append(item)
        return values

    @staticmethod
    def p(i, sample_size, weights):
        """
        Given a weighted set and sample size return the probabilty that the
        weight `i` will be present in the sample.

        Created to test the output of the `SomeOf` maker class. The math was
        provided by Andy Blackshaw - thank you dad :)
        """

        # Determine the initial pick values
        weight_i = weights[i]
        weights_sum = sum(weights)

        # Build a list of weights that don't contain the weight `i`. This list will
        # be used to build the possible picks before weight `i`.
        other_weights = list(weights)
        del other_weights[i]

        # Calculate the probability
        probability_of_i = 0
        for picks in range(0, sample_size):

            # Build the list of possible permutations for this pick in the sample
            permutations = list(itertools.permutations(other_weights, picks))

            # Calculate the probability for this permutation
            permutation_probabilities = []
            for permutation in permutations:

                # Calculate the probability for each pick in the permutation
                pick_probabilities = []
                pick_weight_sum = weights_sum

                for pick in permutation:
                    pick_probabilities.append(pick / pick_weight_sum)

                    # Each time we pick we update the sum of the weight the next
                    # pick is from.
                    pick_weight_sum -= pick

                # Add the probability of picking i as the last pick
                pick_probabilities += [weight_i / pick_weight_sum]

                # Multiply all the probabilities for the permutation together
                permutation_probability = reduce(
                    lambda x, y: x * y, pick_probabilities
                    )
                permutation_probabilities.append(permutation_probability)

            # Add together all the probabilities for all permutations together
            probability_of_i += sum(permutation_probabilities)

        return probability_of_i

    @staticmethod
    def weighted(weights, sample_size, with_replacement=False):
        """
        Return a set of random integers 0 <= N <= len(weights) - 1, where the
        weights determine the probability of each possible integer in the set.
        """
        assert sample_size <= len(weights), "The sample size must be smaller \
than or equal to the number of weights it's taken from."

        # Convert weights to floats
        weights = [float(w) for w in weights]
        weight_indexes = list(range(0, len(weights)))

        samples = []
        while len(samples) < sample_size:
            # Choice a weight
            sample = OneOf.weighted(weights)

            # Add the choosen weight to our samples
            samples.append(weight_indexes[sample])

            if not with_replacement:
                # Remove the weight from the list of weights we can select from
                del weights[sample]
                del weight_indexes[sample]

        return samples





from collections import OrderedDict
import re

from mongoframes.factory import blueprints
from mongoframes.factory import makers

from tests.fixtures import *


def test_blueprint_defaults():
    """
    Defining a `Blueprint` class should provide some defaults.
    """

    class DragonBlueprint(blueprints.Blueprint):

        _frame_cls = Dragon

    assert DragonBlueprint._frame_cls == Dragon
    assert DragonBlueprint._instructions == {}
    assert DragonBlueprint._meta_fields == set([])

def test_blueprint_getters():
    """
    The `Blueprint` getter methods:

    - `get_frame_cls`
    - `get_instructions`
    - `get_meta_fields`

    Should return correct values.
    """

    # Configure the blueprint
    class DragonBlueprint(blueprints.Blueprint):

        _frame_cls = Dragon
        _meta_fields = {'dummy_prop'}

        name = makers.Static('Burt')
        breed = makers.Static('Fire-drake')
        dummy_prop = makers.Static('foo')

    # Check the getter return values
    assert DragonBlueprint.get_frame_cls() == Dragon
    assert DragonBlueprint.get_instructions() == OrderedDict([
        ('name', DragonBlueprint.name),
        ('breed', DragonBlueprint.breed),
        ('dummy_prop', DragonBlueprint.dummy_prop)
        ])
    assert DragonBlueprint.get_meta_fields() == {'dummy_prop'}

def test_blueprint_assemble():
    """
    The `Blueprint.assemble` method should return an assembled a
    document/dictionary for the blueprint.
    """

    # Configure the blueprint
    class DragonBlueprint(blueprints.Blueprint):

        _frame_cls = Dragon
        _meta_fields = {'dummy_prop'}

        name = makers.Static('Burt')
        breed = makers.Static('Fire-drake')
        dummy_prop = makers.Static('foo')

    # Check the assembled output of the blueprint is as expected
    assembled = DragonBlueprint.assemble()

    assert assembled == {
        'breed': 'Fire-drake',
        'dummy_prop': 'foo',
        'name': 'Burt'
        }

def test_blueprint_finish():
    """
    The `Blueprint.finish` method should return a finished document/dictionary
    and meta document/dictionary for the blueprint.
    """

    # Configure the blueprint
    class DragonBlueprint(blueprints.Blueprint):

        _frame_cls = Dragon
        _meta_fields = {'dummy_prop'}

        name = makers.Static('Burt')
        breed = makers.Static('Fire-drake')
        dummy_prop = makers.Static('foo')

    # Check the finished output of the blueprint is as expected
    finished = DragonBlueprint.finish(DragonBlueprint.assemble())

    assert finished == {
        'breed': 'Fire-drake',
        'dummy_prop': 'foo',
        'name': 'Burt'
        }

def test_blueprint_reassemble():
    """
    The `Blueprint.reassemble` method should reassemble the given fields in a
    preassembled document/dictionary for the blueprint.
    """

    # Configure the blueprint
    class DragonBlueprint(blueprints.Blueprint):

        _frame_cls = Dragon
        _meta_fields = {'dummy_prop'}

        name = makers.Static('Burt')
        breed = makers.Static('Fire-drake')
        dummy_prop = makers.Static('foo')

    # Check the assembled output of the blueprint is as expected
    assembled = DragonBlueprint.assemble()

    # Re-configure the blueprint
    class DragonBlueprint(blueprints.Blueprint):

        _frame_cls = Dragon
        _meta_fields = {'dummy_prop'}

        name = makers.Static('Fred')
        breed = makers.Static('Cold-drake')
        dummy_prop = makers.Static('bar')

    # Check the reassembled output for the blueprint is as expected
    DragonBlueprint.reassemble({'breed', 'name'}, assembled)

    assert assembled == {
        'breed': 'Cold-drake',
        'dummy_prop': 'foo',
        'name': 'Fred'
        }

def test_blueprint_reset(mocker):
    """
    The `Blueprint.reset` method should call the reset of all makers in the
    blueprints instructions.
    """

    # Configure the blueprint
    class DragonBlueprint(blueprints.Blueprint):

        _frame_cls = Dragon
        _meta_fields = {'dummy_prop'}

        name = makers.Static('Burt')
        breed = makers.Static('Fire-drake')
        dummy_prop = makers.Static('foo')

    # Spy on the maker reset methods
    mocker.spy(DragonBlueprint._instructions['name'], 'reset')
    mocker.spy(DragonBlueprint._instructions['breed'], 'reset')
    mocker.spy(DragonBlueprint._instructions['dummy_prop'], 'reset')

    # Reset the blueprint
    DragonBlueprint.reset()

    # Check each maker reset method was called
    assert DragonBlueprint._instructions['name'].reset.call_count == 1
    assert DragonBlueprint._instructions['breed'].reset.call_count == 1
    assert DragonBlueprint._instructions['dummy_prop'].reset.call_count == 1


import math
import random

from mongoframes.factory import makers
from mongoframes.factory import quotas
from mongoframes.factory.makers import selections as selection_makers
from mongoframes.queries import Q

from tests.fixtures import *


def test_cycle():
    """
    `Cycle` makers should return values from a list of items (python types of or
    makers) one after another.
    """

    # Configured to cycle through a list of python types
    maker = selection_makers.Cycle(['foo', 'bar', 'zee'])

    index = 0
    for value in ['foo', 'bar', 'zee', 'foo', 'bar']:

        # Check the assembled result
        assembled = maker._assemble()
        assert assembled == [index, None]

        # Check the finished result
        finished = maker._finish(assembled)
        assert finished == value

        # Track the index
        index += 1
        if index >= 3:
            index = 0

    # Configured to cycle throguh a list of makers
    maker = selection_makers.Cycle([
        makers.Static('foo'),
        makers.Static('bar'),
        makers.Static('zee')
        ])

    index = 0
    for value in ['foo', 'bar', 'zee', 'foo', 'bar']:

        # Check the assembled result
        assembled = maker._assemble()
        assert assembled == [index, value]

        # Check the finished result
        finished = maker._finish(assembled)
        assert finished == value

        # Track the index
        index += 1
        if index >= 3:
            index = 0

    # Reset should reset the cycle to the start
    maker.reset()

    # Check the assembled result
    assembled = maker._assemble()
    assert assembled == [0, 'foo']

    # Check the finished result
    finished = maker._finish(assembled)
    assert finished == 'foo'

def test_one_of():
    """
    `OneOf` makers should return one value at random (optionally weighted) from
    a list of items (python types of or makers).
    """

    # Seed the random generator to ensure test results are consistent
    random.seed(110679)

    # Configured to return an item from a list of python types
    maker = selection_makers.OneOf(['foo', 'bar', 'zee'])

    counts = {'foo': 0 , 'bar': 0, 'zee': 0}
    for i in range(0, 1000):
        # Check the assembled result
        assembled = maker._assemble()
        assert assembled[0] in [0, 1, 2] and assembled[1] == None

        # Check the finished result
        finished = maker._finish(assembled)
        assert finished in ['foo', 'bar', 'zee']

        # Count the occurances
        counts[finished] += 1

    # Confirm the counts are approx. evenly distributed
    for value in ['foo', 'bar', 'zee']:
        assert int(round(counts[value] / 100)) == 3

    # Configured to return an item from a list of makers
    maker = selection_makers.OneOf([
        makers.Static('foo'),
        makers.Static('bar'),
        makers.Static('zee')
        ])

    counts = {'foo': 0 , 'bar': 0, 'zee': 0}
    for i in range(0, 1000):
        # Check the assembled result
        assembled = maker._assemble()
        assert assembled[0] in [0, 1, 2]
        assert assembled[1] in ['foo', 'bar', 'zee']

        # Check the finished result
        finished = maker._finish(assembled)
        assert finished in ['foo', 'bar', 'zee']

        # Count the occurances
        counts[finished] += 1

    # Confirm the counts are approx. evenly distributed
    for value in ['foo', 'bar', 'zee']:
        assert int(round(counts[value] / 100)) == 3

    # Configured to return using a weighted bias
    maker = selection_makers.OneOf(
        ['foo', 'bar', 'zee'],
        [quotas.Quota(10), quotas.Quota(30), quotas.Quota(60)]
        )

    counts = {'foo': 0 , 'bar': 0, 'zee': 0}
    for i in range(0, 1000):
        # Check the assembled result
        assembled = maker._assemble()
        assert assembled[0] in [0, 1, 2] and assembled[1] == None

        # Check the finished result
        finished = maker._finish(assembled)
        assert finished in ['foo', 'bar', 'zee']

        # Count the occurances
        counts[finished] += 1

    # Confirm the counts are approx. distributed based on the weights
    assert int(round(counts['foo'] / 100)) == 1
    assert int(round(counts['bar'] / 100)) == 3
    assert int(round(counts['zee'] / 100)) == 6

def test_random_reference(mongo_client, example_dataset_many):
    """
    `RandomReference` makers should return the Id of a document from from the
    specified `Frame`s collection at random.
    """

    # Seed the random generator to ensure test results are consistent
    random.seed(110679)

    # Configured without a constraint
    maker = selection_makers.RandomReference(ComplexDragon)

    # Check the assembled result
    assembled = maker._assemble()
    assert math.floor(assembled * 100) == 77

    # Check the finished result
    finished = maker._finish(assembled)
    assert finished == ComplexDragon.one(Q.name == 'Albert')._id

    # Configured with a constraint
    maker = selection_makers.RandomReference(ComplexDragon, Q.name != 'Albert')

    # Check the assembled result
    assembled = maker._assemble()
    assert math.floor(assembled * 100) == 78

    # Check the finished result
    finished = maker._finish(assembled)
    assert finished == ComplexDragon.one(Q.name == 'Fred')._id

def test_some_of():
    """
    `SomeOf` makers should return a list of values at random (optionally
    weighted) from a list of items (python types of or makers).
    """

    # Seed the random generator to ensure test results are consistent
    random.seed(110679)

    # Define the choices we'll be sampling from
    choices = ['foo', 'bar', 'zee', 'oof', 'rab', 'eez']
    choices_range = range(0, len(choices))
    choices_set = set(choices)

    # Configured to return a sample from a list of python types
    maker = selection_makers.SomeOf(list(choices), quotas.Quota(3))

    counts = {c: 0 for c in choices}
    for i in range(0, 1000):
        # Check the assembled result
        assembled = maker._assemble()
        assert len(assembled) == 3
        for item in assembled:
            assert item[0] in choices_range and item[1] == None

        # Check the finished result
        finished = maker._finish(assembled)
        assert len(set(finished)) == 3
        assert set(finished).issubset(choices_set)

        # Count occurances
        for value in finished:
            counts[value] += 1

    # Confirm the counts are approx. evenly distributed
    for value in choices:
        assert int(round(counts[value] / 100)) == 5

    # Configured to return a sample from a list of makers
    maker = selection_makers.SomeOf(
        [makers.Static(c) for c in choices],
        quotas.Quota(3)
        )

    counts = {c: 0 for c in choices}
    for i in range(0, 1000):
        # Check the assembled result
        assembled = maker._assemble()
        assert len(assembled) == 3
        for item in assembled:
            assert item[0] in choices_range and item[1] in choices

        # Check the finished result
        finished = maker._finish(assembled)
        assert len(set(finished)) == 3
        assert set(finished).issubset(choices_set)

        # Count occurances
        for value in finished:
            counts[value] += 1

    # Confirm the counts are approx. evenly distributed
    for value in choices:
        assert int(round(counts[value] / 100)) == 5

    # Configured to return a sample from a list of python types weighted
    maker = selection_makers.SomeOf(
        list(choices),
        quotas.Quota(3),
        weights=[1, 2, 4, 8, 16, 32]
        )

    counts = {c: 0 for c in choices}
    for i in range(0, 1000):
        # Check the assembled result
        assembled = maker._assemble()
        assert len(assembled) == 3
        for item in assembled:
            assert item[0] in choices_range and item[1] == None

        # Check the finished result
        finished = maker._finish(assembled)
        assert len(finished) == 3
        assert set(finished).issubset(choices_set)

        # Count occurances
        for value in finished:
            counts[value] += 1

    # Confirm the counts are approx. based on the weights
    for i, value in enumerate(choices):

        count = counts[value] / 1000
        prob = maker.p(i, 3, [1, 2, 4, 8, 16, 32])
        tol = prob * 0.15

        assert count > (prob - tol) and count < (prob + tol)

    # Configured to return a sample from a list of python types with replacement
    maker = selection_makers.SomeOf(
        [makers.Static(c) for c in choices],
        quotas.Quota(3),
        with_replacement=False
        )

    not_uniques = 0
    for i in range(0, 1000):
        # Check the assembled result
        assembled = maker._assemble()
        assert len(assembled) == 3
        for item in assembled:
            assert item[0] in choices_range and item[1] in choices

        # Check the finished result
        finished = maker._finish(assembled)
        assert len(finished) == 3
        assert set(finished).issubset(choices_set)

        # Count occurances of values with non-unique values
        if len(set(value)) < 3:
            not_uniques += 1

    # Check that some values where generated with non-unique items
    assert not_uniques > 0

    # Configured to return a sample from a list of python types weighted with
    # replacement.
    maker = selection_makers.SomeOf(
        list(choices),
        quotas.Quota(3),
        weights=[1, 2, 4, 8, 16, 32],
        with_replacement=True
        )

    counts = {c: 0 for c in choices}
    for i in range(0, 1000):
        # Check the assembled result
        assembled = maker._assemble()
        assert len(assembled) == 3
        for item in assembled:
            assert item[0] in choices_range and item[1] == None

        # Check the finished result
        finished = maker._finish(assembled)
        assert len(finished) == 3
        assert set(finished).issubset(choices_set)

        # Count occurances
        for value in finished:
            counts[value] += 1

    # Confirm the counts are approx. (+/- 15% tolerance) based on the weights
    weight = 1
    for value in choices:
        count = counts[value]
        prob = (weight / 63.0) * 3000.0
        tol = prob * 0.15

        assert count > (prob - tol) and count < (prob + tol)

        weight *= 2


from urllib.parse import urlencode

from mongoframes.factory import quotas
from mongoframes.factory.makers import images as image_makers

from tests.fixtures import *


def test_image_url():
    """
    `ImageURL` makers should return the URL for a placeholder image service such
    as `http://fakeimg.pl`.
    """

    # Generate an image URL for the default service provider
    maker = image_makers.ImageURL(
        quotas.Quota(100),
        quotas.Quota(200),
        background='DDDDDD',
        foreground='888888',
        options={'text': 'foo'}
        )

    # Check the assembled result
    image_url = '//fakeimg.pl/100x200/DDDDDD/888888/?text=foo'
    assembled = maker._assemble()
    assert assembled == image_url

    # Check the finished result
    finished = maker._finish(assembled)
    assert finished == image_url

    # Configured for a custom service provider
    def placehold_it_formatter(
        service_url,
        width,
        height,
        background,
        foreground,
        options
        ):
        """Generage an image URL for the placehold.it service"""

        # Build the base URL
        image_tmp = '{service_url}/{width}x{height}/{background}/{foreground}'
        image_url = image_tmp.format(
            service_url=service_url,
            width=width,
            height=height,
            background=background,
            foreground=foreground
            )

        # Check for a format option
        fmt = ''
        if 'format' in options:
            fmt = options.pop('format')
            image_url += '.' + fmt

        # Add any other options
        if options:
            image_url += '?' + urlencode(options)

        return image_url

    maker = image_makers.ImageURL(
        quotas.Quota(100),
        quotas.Quota(200),
        background='DDDDDD',
        foreground='888888',
        options={'text': 'foo', 'format': 'png'},
        service_url='//placehold.it',
        service_formatter=placehold_it_formatter
        )

    # Check the assembled result
    image_url = '//placehold.it/100x200/DDDDDD/888888.png?text=foo'
    assembled = maker._assemble()
    assert assembled == image_url

    # Check the finished result
    finished = maker._finish(assembled)
    assert finished == image_url


from blinker import signal
from bson.objectid import ObjectId
from copy import deepcopy
from datetime import date, datetime, time, timezone

__all__ = [
    'Frame',
    'SubFrame'
    ]


class _BaseFrame:
    """
    Base class for Frames and SubFrames.
    """

    # A cache of key lists generated from path strings used for performance (see
    # `_path_to_keys`.
    _path_to_keys_cache = {}

    def __init__(self, *args, **kwargs):
        # Set the document against the frame (or assign an empty one if one
        # isn't provided).
        if args:
            self._document = args[0]
        else:
            self._document = kwargs

    # Get/Set attribute methods are overwritten to support for setting values
    # against the `_document`. Attribute names are converted to camelcase.

    def __getattr__(self, name):
        if '_document' in self.__dict__ and name in self._fields:
            return self.__dict__['_document'].get(name, None)
        raise AttributeError(
            "'{0}' has no attribute '{1}'".format(self.__class__.__name__, name)
            )

    def __setattr__(self, name, value):
        if '_document' in self.__dict__ and name in self._fields:
            self.__dict__['_document'][name] = value
        else:
            super(_BaseFrame, self).__setattr__(name, value)

    def __getitem__(self, name):
        return self.__dict__['_document'][name]

    def __contains__(self, name):
        return name in self.__dict__['_document']

    def get(self, name, default=None):
        return self.__dict__['_document'].get(name, default)

    # Serializing

    def to_json_type(self):
        """
        Return a dictionary for the document with values converted to JSON safe
        types.
        """
        document_dict = self._json_safe(self._document)
        self._remove_keys(document_dict, self._private_fields)
        return document_dict

    @classmethod
    def _json_safe(cls, value):
        """Return a JSON safe value"""
        # Date
        if type(value) == date:
            return str(value)

        # Datetime
        elif type(value) == datetime:
            return value.strftime('%Y-%m-%d %H:%M:%S')

        # Object Id
        elif isinstance(value, ObjectId):
            return str(value)

        # Frame
        elif isinstance(value, _BaseFrame):
            return value.to_json_type()

        # Lists
        elif isinstance(value, (list, tuple)):
            return [cls._json_safe(v) for v in value]

        # Dictionaries
        elif isinstance(value, dict):
            return {k:cls._json_safe(v) for k, v in value.items()}

        return value

    @classmethod
    def _path_to_keys(cls, path):
        """Return a list of keys for a given path"""

        # Paths are cached for performance
        keys = _BaseFrame._path_to_keys_cache.get(path)
        if keys is None:
            keys = _BaseFrame._path_to_keys_cache[path] = path.split('.')

        return keys

    @classmethod
    def _path_to_value(cls, path, parent_dict):
        """Return a value from a dictionary at the given path"""
        keys = cls._path_to_keys(path)

        # Traverse to the tip of the path
        child_dict = parent_dict
        for key in keys[:-1]:
            child_dict = child_dict.get(key)
            if child_dict is None:
                return

        return child_dict.get(keys[-1])

    @classmethod
    def _remove_keys(cls, parent_dict, paths):
        """
        Remove a list of keys from a dictionary.

        Keys are specified as a series of `.` separated paths for keys in child
        dictionaries, e.g 'parent_key.child_key.grandchild_key'.
        """

        for path in paths:
            keys = cls._path_to_keys(path)

            # Traverse to the tip of the path
            child_dict = parent_dict
            for key in keys[:-1]:
                child_dict = child_dict.get(key)

                if child_dict is None:
                    break

            if child_dict is None:
                continue

            # Remove the key
            if keys[-1] in child_dict:
                child_dict.pop(keys[-1])

    # Public methods

    @classmethod
    def get_fields(cls):
        """Return the set of fields defined for the class"""
        return set(cls._fields)

    @classmethod
    def get_private_fields(cls):
        """Return the set of private fields defined for the class"""
        return set(cls._private_fields)


class _FrameMeta(type):
    """
    Meta class for `Frame`s to ensure an `_id` is present in any defined set of
    fields.
    """

    def __new__(meta, name, bases, dct):

        # If a set of fields is defined ensure it contains `_id`
        if '_fields' in dct and not '_id' in dct['_fields']:
            dct['_fields'].update({'_id'})

        # If no collection name is set then use the class name
        if dct.get('_collection') is None:
            dct['_collection'] = name

        return super(_FrameMeta, meta).__new__(meta, name, bases, dct)


class Frame(_BaseFrame, metaclass=_FrameMeta):
    """
    Frames allow documents to be wrapped in a class adding support for dot
    notation access to attributes and numerous short-cut/helper methods.
    """

    # The MongoDB client used to interface with the database
    _client = None

    # The database on which this collection the class represents is located
    _db = None

    # The database collection this class represents
    _collection = None

    # The documents defined fields
    _fields = set()

    # A set of private fields that will be excluded from the output of
    # `to_json_type`.
    _private_fields = set()

    # Default projection
    _default_projection = None

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self._id == other._id

    def __hash__(self):
        if not self._id:
            raise TypeError('Cannot hash a document without an `_id` set.')
        return int(str(self._id), 16)

    def __lt__(self, other):
        return self._id < other._id

    # Operations

    def insert(self):
        """Insert this document"""
        from mongoframes.queries import to_refs

        # Send insert signal
        signal('insert').send(self.__class__, frames=[self])

        # Prepare the document to be inserted
        document = to_refs(self._document)

        # Insert the document and update the Id
        self._id = self.get_collection().insert_one(document).inserted_id

        # Send inserted signal
        signal('inserted').send(self.__class__, frames=[self])

    def update(self, *fields):
        """
        Update this document. Optionally a specific list of fields to update can
        be specified.
        """
        from mongoframes.queries import to_refs

        assert '_id' in self._document, "Can't update documents without `_id`"

        # Send update signal
        signal('update').send(self.__class__, frames=[self])

        # Check for selective updates
        if len(fields) > 0:
            document = {}
            for field in fields:
                document[field] = self._path_to_value(field, self._document)
        else:
            document = self._document

        # Prepare the document to be updated
        document = to_refs(document)
        document.pop('_id', None)

        # Update the document
        self.get_collection().update_one({'_id': self._id}, {'$set': document})

        # Send updated signal
        signal('updated').send(self.__class__, frames=[self])

    def upsert(self, *fields):
        """
        Update or Insert this document depending on whether it exists or not.
        The presense of an `_id` value in the document is used to determine if
        the document exists.

        NOTE: This method is not the same as specifying the `upsert` flag when
        calling MongoDB. When called for a document with an `_id` value, this
        method will call the database to see if a record with that Id exists,
        if not it will call `insert`, if so it will call `update`. This
        operation is therefore not atomic and much slower than the equivalent
        MongoDB operation (due to the extra call).
        """

        # If no `_id` is provided then we insert the document
        if not self._id:
            return self.insert()

        # If an `_id` is provided then we need to check if it exists before
        # performing the `upsert`.
        #
        if self.count({'_id': self._id}) == 0:
            self.insert()
        else:
            self.update(*fields)

    def delete(self):
        """Delete this document"""

        assert '_id' in self._document, "Can't delete documents without `_id`"

        # Send delete signal
        signal('delete').send(self.__class__, frames=[self])

        # Delete the document
        self.get_collection().delete_one({'_id': self._id})

        # Send deleted signal
        signal('deleted').send(self.__class__, frames=[self])

    @classmethod
    def insert_many(cls, documents):
        """Insert a list of documents"""
        from mongoframes.queries import to_refs

        # Ensure all documents have been converted to frames
        frames = cls._ensure_frames(documents)

        # Send insert signal
        signal('insert').send(cls, frames=frames)

        # Prepare the documents to be inserted
        documents = [to_refs(f._document) for f in frames]

        # Bulk insert
        ids = cls.get_collection().insert_many(documents).inserted_ids

        # Apply the Ids to the frames
        for i, id in enumerate(ids):
            frames[i]._id = id

        # Send inserted signal
        signal('inserted').send(cls, frames=frames)

        return frames

    @classmethod
    def update_many(cls, documents, *fields):
        """
        Update multiple documents. Optionally a specific list of fields to
        update can be specified.
        """
        from mongoframes.queries import to_refs

        # Ensure all documents have been converted to frames
        frames = cls._ensure_frames(documents)

        all_count = len(documents)
        assert len([f for f in frames if '_id' in f._document]) == all_count, \
                "Can't update documents without `_id`s"

        # Send update signal
        signal('update').send(cls, frames=frames)

        # Prepare the documents to be updated

        # Check for selective updates
        if len(fields) > 0:
            documents = []
            for frame in frames:
                document = {'_id': frame._id}
                for field in fields:
                    document[field] = cls._path_to_value(
                        field,
                        frame._document
                        )
                documents.append(to_refs(document))
        else:
            documents = [to_refs(f._document) for f in frames]

        # Update the documents
        for document in documents:
            _id = document.pop('_id')
            cls.get_collection().update(
                {'_id': _id}, {'$set': document})

        # Send updated signal
        signal('updated').send(cls.__class__, frames=frames)

    @classmethod
    def delete_many(cls, documents):
        """Delete multiple documents"""

        # Ensure all documents have been converted to frames
        frames = cls._ensure_frames(documents)

        all_count = len(documents)
        assert len([f for f in frames if '_id' in f._document]) == all_count, \
                "Can't delete documents without `_id`s"

        # Send delete signal
        signal('delete').send(cls, frames=frames)

        # Prepare the documents to be deleted
        ids = [f._id for f in frames]

        # Delete the documents
        cls.get_collection().delete_many({'_id': {'$in': ids}})

        # Send deleted signal
        signal('deleted').send(cls.__class__, frames=frames)

    @classmethod
    def _ensure_frames(cls, documents):
        """
        Ensure all items in a list are frames by converting those that aren't.
        """
        frames = []
        for document in documents:
            if not isinstance(document, Frame):
                frames.append(cls(document))
            else:
                frames.append(document)
        return frames

    # Querying

    def reload(self, **kwargs):
        """Reload the document"""
        frame = self.one({'_id': self._id}, **kwargs)
        self._document = frame._document

    @classmethod
    def by_id(cls, id, **kwargs):
        """Get a document by ID"""
        return cls.one({'_id': id}, **kwargs)

    @classmethod
    def count(cls, filter=None, **kwargs):
        """Return a count of documents matching the filter"""
        from mongoframes.queries import Condition, Group, to_refs

        if isinstance(filter, (Condition, Group)):
            filter = filter.to_dict()

        return cls.get_collection().count(to_refs(filter), **kwargs)

    @classmethod
    def ids(cls, filter=None, **kwargs):
        """Return a list of Ids for documents matching the filter"""
        from mongoframes.queries import Condition, Group, to_refs

        # Find the documents
        if isinstance(filter, (Condition, Group)):
            filter = filter.to_dict()

        documents = cls.get_collection().find(
            to_refs(filter),
            projection={'_id': True},
            **kwargs
            )

        return [d['_id'] for d in list(documents)]

    @classmethod
    def one(cls, filter=None, **kwargs):
        """Return the first document matching the filter"""
        from mongoframes.queries import Condition, Group, to_refs

        # Flatten the projection
        kwargs['projection'], references, subs = \
                cls._flatten_projection(
                    kwargs.get('projection', cls._default_projection)
                    )

        # Find the document
        if isinstance(filter, (Condition, Group)):
            filter = filter.to_dict()

        document = cls.get_collection().find_one(to_refs(filter), **kwargs)

        # Make sure we found a document
        if not document:
            return

        # Dereference the document (if required)
        if references:
            cls._dereference([document], references)

        # Add sub-frames to the document (if required)
        if subs:
            cls._apply_sub_frames([document], subs)

        return cls(document)

    @classmethod
    def many(cls, filter=None, **kwargs):
        """Return a list of documents matching the filter"""
        from mongoframes.queries import Condition, Group, to_refs

        # Flatten the projection
        kwargs['projection'], references, subs = \
                cls._flatten_projection(
                    kwargs.get('projection', cls._default_projection)
                    )

        # Find the documents
        if isinstance(filter, (Condition, Group)):
            filter = filter.to_dict()

        documents = list(cls.get_collection().find(to_refs(filter), **kwargs))

        # Dereference the documents (if required)
        if references:
            cls._dereference(documents, references)

        # Add sub-frames to the documents (if required)
        if subs:
            cls._apply_sub_frames(documents, subs)

        return [cls(d) for d in documents]

    @classmethod
    def _apply_sub_frames(cls, documents, subs):
        """Convert embedded documents to sub-frames for one or more documents"""

        # Dereference each reference
        for path, projection in subs.items():

            # Get the SubFrame class we'll use to wrap the embedded document
            sub = None
            expect_map = False
            if '$sub' in projection:
                sub = projection.pop('$sub')
            elif '$sub.' in projection:
                sub = projection.pop('$sub.')
                expect_map = True
            else:
                continue

            # Add sub-frames to the documents
            raw_subs = []
            for document in documents:
                value = cls._path_to_value(path, document)
                if not value:
                    continue

                if isinstance(value, dict):
                    if expect_map:
                        # Dictionary of embedded documents
                        raw_subs += value.values()
                        for k, v in value.items():
                            if isinstance(v ,list):
                                value[k] = [
                                    sub(u) for u in v if isinstance(u, dict)]
                            else:
                                value[k] = sub(v)

                    # Single embedded document
                    else:
                        raw_subs.append(value)
                        value = sub(value)

                elif isinstance(value, list):
                    # List of embedded documents
                    raw_subs += value
                    value = [sub(v) for v in value if isinstance(v, dict)]

                else:
                    raise TypeError('Not a supported sub-frame type')

                child_document = document
                keys = cls._path_to_keys(path)
                for key in keys[:-1]:
                    child_document = child_document[key]
                child_document[keys[-1]] = value

            # Apply the projection to the list of sub frames
            if projection:
                sub._apply_projection(raw_subs, projection)

    @classmethod
    def _dereference(cls, documents, references):
        """Dereference one or more documents"""

        # Dereference each reference
        for path, projection in references.items():

            # Check there is a $ref in the projection, else skip it
            if '$ref' not in projection:
                continue

            # Collect Ids of documents to dereference
            ids = set()
            for document in documents:
                value = cls._path_to_value(path, document)
                if not value:
                    continue

                if isinstance(value, list):
                    ids.update(value)

                elif isinstance(value, dict):
                    ids.update(value.values())

                else:
                    ids.add(value)

            # Find the referenced documents
            ref = projection.pop('$ref')
            frames = ref.many(
                {'_id': {'$in': list(ids)}},
                projection=projection
                )
            frames = {f._id: f for f in frames}

            # Add dereferenced frames to the document
            for document in documents:
                value = cls._path_to_value(path, document)
                if not value:
                    continue

                if isinstance(value, list):
                    # List of references
                    value = [frames[id] for id in value if id in frames]

                elif isinstance(value, dict):
                    # Dictionary of references
                    value = {key: frames.get(id) for key, id in value.items()}

                else:
                    value = frames.get(value, None)

                child_document = document
                keys = cls._path_to_keys(path)
                for key in keys[:-1]:
                    child_document = child_document[key]
                child_document[keys[-1]] = value

    @classmethod
    def _flatten_projection(cls, projection):
        """
        Flatten a structured projection (structure projections support for
        projections of (to be) dereferenced fields.
        """

        # If `projection` is empty return a full projection based on `_fields`
        if not projection:
            return {f: True for f in cls._fields}, {}, {}

        # Flatten the projection
        flat_projection = {}
        references = {}
        subs = {}
        inclusive = True
        for key, value in deepcopy(projection).items():
            if isinstance(value, dict):
                # Store a reference/sub-frame projection
                if '$ref' in value:
                    references[key] = value
                elif '$sub' in value or '$sub.' in value:
                    subs[key] = value
                flat_projection[key] = True

            elif key == '$ref':
                # Strip any $ref key
                continue

            elif key == '$sub' or key == '$sub.':
                # Strip any $sub key
                continue

            else:
                # Store the root projection value
                flat_projection[key] = value
                inclusive = False

        # If only references and sub-frames where specified in the projection
        # then return a full projection based on `_fields`.
        if inclusive:
            flat_projection = {f: True for f in cls._fields}

        return flat_projection, references, subs

    # Integrity helpers

    @staticmethod
    def timestamp_insert(sender, frames):
        """
        Timestamp the created and modified fields for all documents. This method
        should be bound to a frame class like so:

        ```
        MyFrameClass.listen('insert', MyFrameClass.timestamp_insert)
        ```
        """
        for frame in frames:
            timestamp = datetime.now(timezone.utc)
            frame.created = timestamp
            frame.modified = timestamp

    @staticmethod
    def timestamp_update(sender, frames):
        """
        Timestamp the modified field for all documents. This method should be
        bound to a frame class like so:

        ```
        MyFrameClass.listen('update', MyFrameClass.timestamp_update)
        ```
        """
        for frame in frames:
            frame.modified = datetime.now(timezone.utc)

    @classmethod
    def cascade(cls, ref_cls, field, frames):
        """Apply a cascading delete (does not emit signals)"""
        from mongoframes.queries import to_refs
        ids = [to_refs(f[field]) for f in frames if f.get(field)]
        ref_cls.get_collection().delete_many({'_id': {'$in': ids}})

    @classmethod
    def nullify(cls, ref_cls, field, frames):
        """Nullify a reference field (does not emit signals)"""
        from mongoframes.queries import to_refs
        ids = [to_refs(f) for f in frames]
        ref_cls.get_collection().update_many(
            {field: {'$in': ids}},
            {'$set': {field: None}}
            )

    @classmethod
    def pull(cls, ref_cls, field, frames):
        """Pull references from a list field (does not emit signals)"""
        from mongoframes.queries import to_refs
        ids = [to_refs(f) for f in frames]
        ref_cls.get_collection().update_many(
            {field: {'$in': ids}},
            {'$pull': {field: {'$in': ids}}}
            )

    # Signals

    @classmethod
    def listen(cls, event, func):
        """Add a callback for a signal against the class"""
        signal(event).connect(func, sender=cls)

    @classmethod
    def stop_listening(cls, event, func):
        """Remove a callback for a signal against the class"""
        signal(event).disconnect(func, sender=cls)

    # Misc.

    @classmethod
    def get_collection(cls):
        """Return a reference to the database collection for the class"""
        return getattr(cls.get_db(), cls._collection)

    @classmethod
    def get_db(cls):
        """Return the database for the collection"""
        if cls._db:
            return getattr(cls._client, cls._db)
        return cls._client.get_default_database()


class SubFrame(_BaseFrame):
    """
    Sub-frames allow embedded documents to be wrapped in a class adding support
    for dot notation access to attributes.
    """

    # The documents defined fields
    _fields = set()

    # A set of private fields that will be excluded from the output of
    # `to_json_type`.
    _private_fields = set()

    @classmethod
    def _apply_projection(cls, documents, projection):

        # Find reference and sub-frame mappings
        references = {}
        subs = {}
        for key, value in deepcopy(projection).items():

            if not isinstance(value, dict):
                continue

            # Store a reference/sub-frame projection
            if '$ref' in value:
                references[key] = value
            elif '$sub' in value or '$sub.' in value:
                subs[key] = value

        # Dereference the documents (if required)
        if references:
            Frame._dereference(documents, references)

        # Add sub-frames to the documents (if required)
        if subs:
            Frame._apply_sub_frames(documents, subs)


import datetime

from mongoframes.factory.makers import dates as date_makers

from tests.fixtures import *


def test_date_between():
    """
    `DateBetween` makers should return a random date between two given dates.
    """

    # Configured between two `datetime.date` instances
    min_date = datetime.date(2016, 1, 1)
    max_date = datetime.date(2016, 6, 1)
    maker = date_makers.DateBetween(min_date, max_date)

    # Calculate the number of days between the two dates
    days = (max_date - min_date).total_seconds() / 86400

    # Check the assembled result
    assembled = maker._assemble()
    assert assembled >= 0 and assembled <= days

    # Check the finished result
    finished = maker._finish(assembled)
    assert finished == min_date + datetime.timedelta(seconds=assembled * 86400)

    # Configured between two string dates
    maker = date_makers.DateBetween('today-5', 'today+5')

    # Check the assembled result
    assembled = maker._assemble()
    assert assembled >= 0 and assembled <= 10

    # Check the finished result
    min_date = datetime.date.today() - datetime.timedelta(seconds=5 * 86400)
    finished = maker._finish(assembled)
    assert finished == min_date + datetime.timedelta(seconds=assembled * 86400)

    # Check we can correctly parse tomorrow, today, yesterday
    maker_cls = date_makers.DateBetween

    # Today
    assert maker_cls.parse_date('today') == datetime.date.today()

    # Tomorrow
    tomorrow = datetime.date.today() + datetime.timedelta(days=1)
    assert maker_cls.parse_date('tomorrow') == tomorrow

    # Yesterday
    yesterday = datetime.date.today() - datetime.timedelta(days=1)
    assert maker_cls.parse_date('yesterday') == yesterday


import contextlib
import random

import faker

from mongoframes.queries import Q

__all__ = [
    'DictOf',
    'Faker',
    'Lambda',
    'ListOf',
    'Reference',
    'Static',
    'SubFactory',
    'Unique'
    ]


class Maker:
    """
    A base class for all Maker classes.
    """

    def __init__(self):

        # The document the maker is assembling/finishing data for
        self._document = None

    def __call__(self, *args):
        if args:
            return self._finish(*args)
        return self._assemble()

    @property
    def document(self):
        # Return the target document
        return self._document

    def reset(self):
        """Reset the maker instance"""
        pass

    def _assemble(self):
        return None

    def _finish(self, value):
        return value

    @contextlib.contextmanager
    def target(self, document):
        self._document = document
        yield
        self._document = None


class DictOf(Maker):
    """
    Make a dictionary of key/values where each value is a set value or generated
    using a maker.
    """

    def __init__(self, table):
        super().__init__()

        # The table of keyword arguments that will be used to generate the
        # dictionary.
        self._table = table

    def _assemble(self):
        table = {}
        for k, v in self._table.items():
            if isinstance(v, Maker):
                table[k] = v._assemble()
            else:
                table[k] = None
        return table

    def _finish(self, value):
        table = {}
        for k, v in self._table.items():
            if isinstance(v, Maker):
                table[k] = self._table[k]._finish(value[k])
            else:
                table[k] = self._table[k]

        return table


class Faker(Maker):
    """
    Use any faker provider to generate a value (see
    http://fake-factory.readthedocs.io/)
    """

    default_locale = 'en_US'

    def __init__(self, provider, assembler=True, locale=None, **kwargs):
        super().__init__()

        # The provider that will be used to generate the value
        self._provider = provider

        # Flag indicating if the providers should be called in _assemble (True)
        # or _finish (False).
        self._assembler = assembler

        # The locale that will be used by the faker factory
        self._locale = locale or self.default_locale

        # The keyword arguments for the provider
        self._kwargs = kwargs

    def _assemble(self):
        if not self._assembler:
            return None
        provider = getattr(self.get_fake(self._locale), self._provider)
        return provider(**self._kwargs)

    def _finish(self, value):
        if self._assembler:
            return value
        provider = getattr(self.get_fake(self._locale), self._provider)
        return provider(**self._kwargs)

    @staticmethod
    def get_fake(locale=None):
        """Return a shared faker factory used to generate fake data"""
        if locale is None:
            locale = Faker.default_locale

        if not hasattr(Maker, '_fake_' + locale):
            Faker._fake = faker.Factory.create(locale)
        return Faker._fake


class Lambda(Maker):
    """
    Use a function to generate a value.
    """

    def __init__(self, func, assembler=True, finisher=False):
        super().__init__()

        assert assembler or finisher, \
                'Either `assembler` or `finisher` must be true for lambda'

        # The function to call
        self._func = func

        # Flag indicating if the lambda function should be called in _assemble
        self._assembler = assembler

        # Flag indicating if the lambda function should be called in _finish
        self._finisher = finisher

    def _assemble(self):
        if self._assembler:
            return self._func(self.document)
        return None

    def _finish(self, value):
        if self._finisher:
            return self._func(self.document, value)
        return value


class ListOf(Maker):
    """
    Generate a list of values of the given quantity using the specified maker.
    """

    def __init__(self, maker, quantity, reset_maker=False):
        super().__init__()

        # The maker used to generate each value in the list
        self._maker = maker

        # The number of list items to generate
        self._quantity = quantity

        # A flag indicating if the maker should be reset each time the list is
        # generated.
        self._reset_maker = reset_maker

    def _assemble(self):
        quantity = int(self._quantity)

        if self._reset_maker:
            self._maker.reset()

        return [self._maker() for i in range(0, quantity)]

    def _finish(self, value):

        if self._reset_maker:
            self._maker.reset()

        return [self._maker(v) for v in value]


class Static(Maker):
    """
    A maker that returns a fixed value.
    """

    def __init__(self, value, assembler=True):
        super().__init__()

        # The value to return
        self._value = value

        # Flag indicating if the value should be return in _assemble (True) or
        # _finish (False).
        self._assembler = assembler

    def _assemble(self):
        if self._assembler:
            return self._value
        return None

    def _finish(self, value):
        if self._assembler:
            return value
        return self._value


class SubFactory(Maker):
    """
    A maker that generates sub-documents.
    """

    def __init__(self, blueprint):
        super().__init__()

        # The blueprint to produce
        self._blueprint = blueprint

    def reset(self):
        """Reset the blueprint for the maker"""
        self._blueprint.reset()

    def _assemble(self):
        return self._blueprint.assemble()

    def _finish(self, value):
        document = self._blueprint.finish(value)

        # Separate out any meta fields
        meta_document = {}
        for field_name in self._blueprint._meta_fields:
            meta_document[field_name] = document[field_name]
            document.pop(field_name)

        # Initialize the sub-frame
        sub_frame = self._blueprint.get_frame_cls()(document)

        # Apply any meta fields
        for key, value in meta_document.items():
            setattr(sub_frame, key, value)

        return sub_frame


class Unique(Maker):
    """
    Ensure that unique values are generated by a maker.
    """

    def __init__(self,
        maker,
        exclude=None,
        assembler=True,
        max_attempts=1000
        ):
        super().__init__()

        # The maker that will generate values
        self._maker = maker

        # A set of existing values
        self._exclude = exclude or set([])

        # Flag indicating if the providers should be called in _assemble (True)
        # or _finish (False).
        self._assembler = assembler

        # A set of used values
        self._used_values = set(self._exclude)

        # The maximum number of attempts to generate unique data that should be
        # performed.
        self._max_attempts = max_attempts

    def reset(self):
        """Reset the set of used values"""
        self._used_values = set(self._exclude)

    def _get_unique(self, *args):
        """Generate a unique value using the assigned maker"""

        # Generate a unique values
        value = ''
        attempts = 0
        while True:
            attempts += 1
            value = self._maker(*args)
            if value not in self._used_values:
                break

            assert attempts < self._max_attempts, \
                'Too many attempts to generate a unique value'

        # Add the value to the set of used values
        self._used_values.add(value)

        return value

    def _assemble(self):
        if not self._assembler:
            return self._maker()
        return self._get_unique()

    def _finish(self, value):
        if self._assembler:
            return self._maker(value)
        return self._get_unique(value)


import copy

__all__ = ['Factory']


class Factory:
    """
    The `Factory` class is responsible for production of fake data for.
    Production of fake data is a two stage process:

    Assembly (see `assemble`)
    :   A `Quota` of documents is assembled based on a `Blueprint`.

        At this stage the documents contain a mixture of static and dynamic
        data. Dynamic data is data that will be transformed during population,
        for example a field might contain a value of `'now,tomorrow'` which on
        population will be converted to a date/time between now and tomorrow.

        Once assembled the generated documents are returned as a list and can be
        either used immediately to populate the database or saved out as a
        template for populating the database in future (for example when
        building a set of test data).

    Finishing and population (see `finish` and `populate`)
    :   A database is populated based on a `Blueprint` and pre-assembled list of
        documents.

        During this stage dynamic data is converted to static data (this process
        is call finishing) and inserted into the database.

        Before populate inserts the finished documents into the database it
        converts each document into a Frame instance and calls the on_fake
        method against the Blueprint, after the documents are inserted into the
        database it calls the on_faked method against the Blueprint.
    """

    # Public methods

    def assemble(self, blueprint, quota):
        """Assemble a quota of documents"""

        # Reset the blueprint
        blueprint.reset()

        # Assemble the documents
        documents = []
        for i in range(0, int(quota)):
            documents.append(blueprint.assemble())

        return documents

    def finish(self, blueprint, documents):
        """Finish a list of pre-assembled documents"""

        # Reset the blueprint
        blueprint.reset()

        # Finish the documents
        finished = []
        for document in documents:
            finished.append(blueprint.finish(document))

        return finished

    def populate(self, blueprint, documents):
        """Populate the database with documents"""

        # Finish the documents
        documents = self.finish(blueprint, documents)

        # Convert the documents to frame instances
        frames = []
        for document in documents:
            # Separate out any meta fields
            meta_document = {}
            for field_name in blueprint._meta_fields:
                meta_document[field_name] = document[field_name]
                document.pop(field_name)

            # Initialize the frame
            frame = blueprint.get_frame_cls()(document)

            # Apply any meta fields
            for key, value in meta_document.items():
                setattr(frame, key, value)

            frames.append(frame)

        # Insert the documents
        blueprint.on_fake(frames)
        frames = blueprint.get_frame_cls().insert_many(frames)
        blueprint.on_faked(frames)

        return frames

    def reassemble(self, blueprint, fields, documents):
        """
        Reassemble the given set of fields for a list of pre-assembed documents.

        NOTE: Reassembly is done in place, since the data you send the method
        should be JSON type safe, if you need to retain the existing document
        it is recommended that you copy them using `copy.deepcopy`.
        """

        # Reset the blueprint
        blueprint.reset()

        # Reassemble the documents
        for document in documents:
            blueprint.reassemble(fields, document)


from mongoframes.frames import *
from mongoframes.pagination import *
from mongoframes.queries import *
from pymongo.collation import Collation
from pymongo import (
    IndexModel,
    ASCENDING as ASC,
    DESCENDING as DESC,
    GEOSPHERE,
    TEXT
)


from mongoframes.factory import Factory
from mongoframes.factory import blueprints
from mongoframes.factory import makers
from mongoframes.factory import quotas

from tests.fixtures import *


def test_factory_assemble():
    """
    The `Factory.assemble` method should return a quota of assembled
    documents/dictionaries for the given blueprint.
    """

    # Configure the blueprint
    class DragonBlueprint(blueprints.Blueprint):

        _frame_cls = Dragon
        _meta_fields = {'dummy_prop'}

        name = makers.Static('Burt')
        breed = makers.Static('Fire-drake')
        dummy_prop = makers.Static('foo')

    # Configure the factory
    factory = Factory()

    # Assemble a list of documents using the factory
    documents = factory.assemble(DragonBlueprint, quotas.Quota(10))

    # Check the assembled output of the factory is as expected
    for document in documents:
        assert document == {
            'breed': 'Fire-drake',
            'dummy_prop': 'foo',
            'name': 'Burt'
            }

def test_factory_finish():
    """
    The `Factory.assemble` method should return a list of finished
    documents/dictionaries and meta documents/dictionaries in the form
    `[(document, meta_document), ...]` from a blueprint and list of preassembled
    documents/dictionaries.
    """

    # Configure the blueprint
    class DragonBlueprint(blueprints.Blueprint):

        _frame_cls = Dragon
        _meta_fields = {'dummy_prop'}

        name = makers.Static('Burt')
        breed = makers.Static('Fire-drake')
        dummy_prop = makers.Static('foo')

    # Configure the factory
    factory = Factory()

    # Assemble a list of documents using the factory
    documents = factory.assemble(DragonBlueprint, quotas.Quota(10))

    # Finish the assembled documents
    documents = factory.finish(DragonBlueprint, documents)

    # Check the assembled output of the factory is as expected
    for document in documents:
        assert document == {
            'breed': 'Fire-drake',
            'name': 'Burt',
            'dummy_prop': 'foo'
            }

def test_factory_populate(mongo_client, mocker):
    """
    The `Factory.populate` method should return populate a database collection
    using using a blueprint and list of preassembled documents/dictionaries.
    """

    # Configure the blueprint
    class DragonBlueprint(blueprints.Blueprint):

        _frame_cls = Dragon
        _meta_fields = {'dummy_prop'}

        name = makers.Static('Burt')
        breed = makers.Static('Fire-drake')
        dummy_prop = makers.Static('foo')

    # Configure the factory
    factory = Factory()

    # Assemble a list of documents using the factory
    documents = factory.assemble(DragonBlueprint, quotas.Quota(10))

    # Add listeners for fake and faked
    Dragon._on_fake = lambda sender, frames: print('qwe')
    Dragon._on_faked = lambda sender, frames: None

    mocker.spy(Dragon, '_on_fake')
    mocker.spy(Dragon, '_on_faked')

    Dragon.listen('fake', Dragon._on_fake)
    Dragon.listen('faked', Dragon._on_faked)

    # Populate the database with our fake documents
    frames = factory.populate(DragonBlueprint, documents)

    # Check each maker reset method was called
    assert Dragon._on_fake.call_count == 1
    assert Dragon._on_faked.call_count == 1

    # Check the frames created
    for frame in frames:
        assert frame._id is not None
        assert frame.name == 'Burt'
        assert frame.breed == 'Fire-drake'
        assert frame.dummy_prop


def test_factory_reassemble():
    """
    The `Blueprint.reassemble` method should reassemble the given fields in a
    list of preassembled documents/dictionaries.
    """

    # Configure the blueprint
    class DragonBlueprint(blueprints.Blueprint):

        _frame_cls = Dragon
        _meta_fields = {'dummy_prop'}

        name = makers.Static('Burt')
        breed = makers.Static('Fire-drake')
        dummy_prop = makers.Static('foo')

    # Configure the factory
    factory = Factory()

    # Assemble a list of documents using the factory
    documents = factory.assemble(DragonBlueprint, quotas.Quota(10))

    # Re-configure the blueprint and factory
    class DragonBlueprint(blueprints.Blueprint):

        _frame_cls = Dragon
        _meta_fields = {'dummy_prop'}

        name = makers.Static('Fred')
        breed = makers.Static('Cold-drake')
        dummy_prop = makers.Static('bar')

    factory = Factory()

    # Reassemble the documents
    factory.reassemble(DragonBlueprint, {'breed', 'name'}, documents)

    # Check the reassembled output of the factory is as expected
    for document in documents:
        assert document == {
            'breed': 'Cold-drake',
            'dummy_prop': 'foo',
            'name': 'Fred'
            }


"""
A set of helpers to simplify the creation of MongoDB queries.
"""

import re

from pymongo import (ASCENDING, DESCENDING)

__all__ = [
    # Queries
    'Q',

    # Operators
    'All',
    'ElemMatch',
    'Exists',
    'In',
    'Not',
    'NotIn',
    'Size',
    'Type',

    # Groups
    'And',
    'Or',
    'Nor',

    # Sorting
    'SortBy',

    # Utils
    'to_refs'
    ]


# Queries

class Condition:
    """
    A query condition of the form `{path: {operator: value}}`.
    """

    def __init__(self, q, value, operator):
        self.q = q
        self.value = value
        self.operator = operator

    def to_dict(self):
        """Return a dictionary suitable for use with pymongo as a filter"""
        if self.operator == '$eq':
            return {self.q: self.value}
        if self.q is None:
            return {self.operator: self.value}
        return {self.q: {self.operator: self.value}}


class QMeta(type):
    """
    Meta-class for query builder.
    """

    def __getattr__(self, name):
        return Q(name)

    def __getitem__(self, name):
        return Q(name)

    def __eq__(self, other):
        return Condition(None, other, '$eq')

    def __ge__(self, other):
        return Condition(None, other, '$gte')

    def __gt__(self, other):
        return Condition(None, other, '$gt')

    def __le__(self, other):
        return Condition(None, other, '$lte')

    def __lt__(self, other):
        return Condition(None, other, '$lt')

    def __ne__(self, other):
        return Condition(None, other, '$ne')


class Q(metaclass=QMeta):
    """
    Start point for the query creation, the Q class is a special type of class
    that's typically initialized by appending an attribute, for example:

        Q.hit_points > 100

    """

    def __init__(self, path):
        self._path = path

    def __eq__(self, other):
        return Condition(self._path, other, '$eq')

    def __ge__(self, other):
        return Condition(self._path, other, '$gte')

    def __gt__(self, other):
        return Condition(self._path, other, '$gt')

    def __le__(self, other):
        return Condition(self._path, other, '$lte')

    def __lt__(self, other):
        return Condition(self._path, other, '$lt')

    def __ne__(self, other):
        return Condition(self._path, other, '$ne')

    def __getattr__(self, name):
        self._path = '{0}.{1}'.format(self._path, name)
        return self

    def __getitem__(self, name):
        self._path = '{0}.{1}'.format(self._path, name)
        return self


# Operators

def All(q, value):
    """
    The All operator selects documents where the value of the field is an list
    that contains all the specified elements.
    """
    return Condition(q._path, to_refs(value), '$all')

def ElemMatch(q, *conditions):
    """
    The ElemMatch operator matches documents that contain an array field with at
    least one element that matches all the specified query criteria.
    """
    new_condition = {}
    for condition in conditions:
        deep_merge(condition.to_dict(), new_condition)

    return Condition(q._path, new_condition, '$elemMatch')

def Exists(q, value):
    """
    When exists is True, Exists matches the documents that contain the field,
    including documents where the field value is null. If exists is False, the
    query returns only the documents that do not contain the field.
    """
    return Condition(q._path, value, '$exists')

def In(q, value):
    """
    The In operator selects the documents where the value of a field equals any
    value in the specified list.
    """
    return Condition(q._path, to_refs(value), '$in')

def Not(condition):
    """
    Not performs a logical NOT operation on the specified condition and selects
    the documents that do not match. This includes documents that do not contain
    the field.
    """
    return Condition(
        condition.q,
        {condition.operator: condition.value},
        '$not'
        )

def NotIn(q, value):
    """
    The NotIn operator selects documents where the field value is not in the
    specified list or the field does not exists.
    """
    return Condition(q._path, to_refs(value), '$nin')

def Size(q, value):
    """
    The Size operator matches any list with the number of elements specified by
    size.
    """
    return Condition(q._path, value, '$size')

def Type(q, value):
    """
    Type selects documents where the value of the field is an instance of the
    specified BSON type.
    """
    return Condition(q._path, value, '$type')


# Groups

class Group:
    """
    The Group class is used as a base class for operators that group together
    two or more conditions.
    """

    operator = ''

    def __init__(self, *conditions):
        self.conditions = conditions

    def to_dict(self):
        """Return a dictionary suitable for use with pymongo as a filter"""
        raw_conditions = []
        for condition in self.conditions:
            if isinstance(condition, (Condition, Group)):
                raw_conditions.append(condition.to_dict())
            else:
                raw_conditions.append(condition)
        return {self.operator: raw_conditions}


class And(Group):
    """
    And performs a logical AND operation on a list of two or more conditions and
    selects the documents that satisfy all the conditions.
    """

    operator = '$and'


class Or(Group):
    """
    The Or operator performs a logical OR operation on a list of two or more
    conditions and selects the documents that satisfy at least one of the
    conditions.
    """

    operator = '$or'


class Nor(Group):
    """
    Nor performs a logical NOR operation on a list of one or more conditions and
    selects the documents that fail all the conditions.
    """

    operator = '$nor'


# Sorting

def SortBy(*qs):
    """Convert a list of Q objects into list of sort instructions"""

    sort = []
    for q in qs:
        if q._path.endswith('.desc'):
            sort.append((q._path[:-5], DESCENDING))
        else:
            sort.append((q._path, ASCENDING))
    return sort


# Utils

def deep_merge(source, dest):
    """
    Deep merges source dict into dest dict.

    This code was taken directly from the mongothon project:
    https://github.com/gamechanger/mongothon/tree/master/mongothon
    """
    for key, value in source.items():
        if key in dest:
            if isinstance(value, dict) and isinstance(dest[key], dict):
                deep_merge(value, dest[key])
                continue
            elif isinstance(value, list) and isinstance(dest[key], list):
                for item in value:
                    if item not in dest[key]:
                        dest[key].append(item)
                continue
        dest[key] = value

def to_refs(value):
    """Convert all Frame instances within the given value to Ids"""
    from mongoframes.frames import Frame, SubFrame

    # Frame
    if isinstance(value, Frame):
        return value._id

    # SubFrame
    elif isinstance(value, SubFrame):
        return to_refs(value._document)

    # Lists
    elif isinstance(value, (list, tuple)):
        return [to_refs(v) for v in value]

    # Dictionaries
    elif isinstance(value, dict):
        return {k: to_refs(v) for k, v in value.items()}

    return value


import re

from mongoframes.factory import blueprints
from mongoframes.factory import makers
from mongoframes.factory import quotas
from mongoframes.factory.makers import selections as selection_makers
from mongoframes.factory.makers import text as text_makers

from tests.fixtures import *


def test_maker():
    """
    The base maker class should provide context for the current target document.
    """

    document = {'foo': 'bar'}
    maker = makers.Maker()

    # Check the target for the maker is correctly set using the `target` context
    # method.
    with maker.target(document):
        assert maker.document == document

    # Once the maker falls out of context check the document has been unset
    assert maker.document == None

def test_dict_of():
    """
    `DictOf` makers should return a dictionary where each key's value is either
    a JSON type value the output of a maker.
    """

    maker = makers.DictOf({
        'json_type': 'foo',
        'maker': makers.Lambda(lambda doc: 'bar')
        })

    # Check the assembled result
    assembled = maker._assemble()
    assert assembled == {
        'json_type': None,
        'maker': 'bar'
    }

    # Check the finished result
    finished = maker._finish(assembled)
    assert finished == {
        'json_type': 'foo',
        'maker': 'bar'
    }

def test_faker():
    """
    `Faker` makers should call a faker library provider and return the output as
    the value.
    """
    am_pm = {'AM', 'PM'}

    # Configured as assembler
    maker = makers.Faker('am_pm')

    # Check the assembled result
    assembled = maker._assemble()
    assert assembled in am_pm

    # Check the finished result
    finished = maker._finish(assembled)
    assert finished in am_pm

    # Configured as finisher
    maker = makers.Faker('am_pm', assembler=False)

    # Check the assembled result
    assembled = maker._assemble()
    assert assembled == None

    # Check the finished result
    finished = maker._finish(assembled)
    assert finished in am_pm

    # Configured with a different locale
    maker = makers.Faker('postcode', locale='en_GB')

    # Check the assembled result resembles a UK postcode
    assembled = maker._assemble()

    assert re.match('(\w+?\d{1,2}).*', assembled) and len(assembled) <= 8

def test_lambda():
    """
    `Lambda` makers should return the output of the function you initialize them
    with.
    """

    # Configured as assembler
    maker = makers.Lambda(lambda doc: 'foo')

    # Check the assembled result
    assembled = maker._assemble()
    assert assembled == 'foo'

    # Check the finished result
    finished = maker._finish(assembled)
    assert finished == 'foo'

    # Configured as finisher
    maker = makers.Lambda(lambda doc, v: 'bar', assembler=False, finisher=True)

    # Check the assembled result
    assembled = maker._assemble()
    assert assembled == None

    # Check the finished result
    finished = maker._finish(assembled)
    assert finished == 'bar'

    # Configured as both an assembler and finisher
    def func(doc, value=None):
        if value:
            return value + 'bar'
        return 'foo'

    maker = makers.Lambda(func, finisher=True)

    # Check the assembled result
    assembled = maker._assemble()
    assert assembled == 'foo'

    # Check the finished result
    finished = maker._finish(assembled)
    assert finished == 'foobar'

def test_list_of():
    """
    `ListOf` makers should return a list of values generated by calling a maker
    multiple times.
    """

    # Configured to not reset sub-maker
    maker = makers.ListOf(
        selection_makers.Cycle(list('abcde')),
        quotas.Quota(6)
        )

    # Check the assembled result
    assembled = maker._assemble()
    assert assembled == [[i, None] for i in [0, 1, 2, 3, 4, 0]]

    # Check the finished result
    finished = maker._finish(assembled)
    assert finished == list('abcdea')

    # Check that calling the maker again continues from where we left off
    assembled = maker._assemble()
    assert assembled == [[i, None] for i in [1, 2, 3, 4, 0, 1]]

    # Configured to reset sub-maker
    maker = makers.ListOf(
        selection_makers.Cycle(list('abcde')),
        quotas.Quota(6),
        reset_maker=True
        )

    # Call the maker twice
    assembled = maker._assemble()
    assembled = maker._assemble()

    # Check the result was reset after the first call
    assert assembled == [[i, None] for i in [0, 1, 2, 3, 4, 0]]

def test_static():
    """`Static` makers should return the value you initialize them with"""

    # Configured as assembler
    value = {'foo': 'bar'}
    maker = makers.Static(value)

    # Check the assembled result
    assembled = maker._assemble()
    assert assembled == value

    # Check the finished result
    finished = maker._finish(assembled)
    assert finished == value

    # Configured as finisher
    value = {'foo': 'bar'}
    maker = makers.Static(value, assembler=False)

    # Check the assembled result
    assembled = maker._assemble()
    assert assembled == None

    # Check the finished result
    finished = maker._finish(assembled)
    assert finished == value

def test_sub_factory(mocker):
    """
    `SubFactory` makers should return a sub-frame/document using a blueprint.
    """

    # Define a blueprint
    class InventoryBlueprint(blueprints.Blueprint):

        _frame_cls = Inventory

        gold = makers.Static(10)
        skulls = makers.Static(100)

    # Configure the maker
    maker = makers.SubFactory(InventoryBlueprint)

    # Check the assembled result
    assembled = maker._assemble()
    assert assembled == {'gold': 10, 'skulls': 100}

    # Check the finished result
    finished = maker._finish(assembled)
    assert isinstance(finished, Inventory)
    assert finished._document == {'gold': 10, 'skulls': 100}

    # Reset should reset the sub factories associated blueprint
    mocker.spy(InventoryBlueprint, 'reset')
    maker.reset()
    assert InventoryBlueprint.reset.call_count == 1

def test_unique():
    """
    `Unique` makers guarentee a unique value is return from the maker they are
    wrapped around.
    """

    # Confifured as assembler
    maker = makers.Unique(makers.Faker('name'))

    # Generate 100 random names
    names = set([])
    for i in range(0, 20):
        assembled = maker._assemble()
        assert assembled not in names
        names.add(assembled)

    # Confifured as finisher
    maker = makers.Unique(makers.Faker('name'), assembler=False)

    # Generate 100 random names
    names = set([])
    for i in range(0, 20):
        finished = maker._finish(maker._assemble())
        assert finished not in names
        names.add(finished)

    # Check that unique will eventually fail if it cannot generate a unique
    # response with a maker.
    maker = makers.Unique(makers.Static('foo'))

    failed = False
    try:
        for i in range(0, 100):
            finished = maker._finish(maker._assemble())
    except AssertionError:
        failed = True

    assert failed

    # Check that we can include a set of initial exluded values
    maker = makers.Unique(
        text_makers.Sequence('test-{index}'),
        exclude={'test-3'}
        )

    names = set([])
    for i in range(0, 9):
        assembled = maker._assemble()
        names.add(assembled)

    assert 'test-3' not in names

    # Reset should clear the generate unique values from the maker and allow
    # those values to be generated again.
    maker = makers.Unique(makers.Static('foo'))

    failed = False
    try:
        for i in range(0, 100):
            finished = maker._finish(maker._assemble())
            maker.reset()

    except AssertionError:
        failed = True

    assert not failed


import math
import random

from mongoframes.factory import quotas


def test_quota():
    """
    The base `Quota` should convert to a quantity identical to the one it is
    initialized with.
    """

    quota = quotas.Quota(5)

    # Check quota converted to an integer
    assert int(quota) == 5

    # Check quota converted to a float
    assert float(quota) == 5.0

def test_gauss():
    """
    The `Gauss` quota should convert to a random quantity using a gaussian
    distribution and a given mean and standard deviation.
    """

    # Seed the random generator to ensure test results are consistent
    random.seed(110679)

    # Configure the quota
    std_dev = 0.1
    mean = 5
    quota = quotas.Gauss(mean, std_dev)

    # Count the number of times a value falls within each deviation
    std_devs = {1: 0, 2: 0, 3: 0}

    for i in range(0, 10000):
        qty = float(quota)

        # Calculate the deviation of the quantity from the mean
        dev = abs(qty - 5)

        # Tally the occurances within each deviation deviations
        if dev <= std_dev:
            std_devs[1] += 1

        if dev <= std_dev * 2:
            std_devs[2] += 1

        if dev <= std_dev * 3:
            std_devs[3] += 1

    # Check the deviations fall within the 68-95-99.7 rule
    assert math.ceil(std_devs[1] / 100) >= 68
    assert math.ceil(std_devs[2] / 100) >= 95
    assert math.ceil(std_devs[3] / 100) >= 99.7

def test_random():
    """
    The `Random` quota should convert to a random quantity between a minimum and
    maximum value.
    """

    min_qty = 5
    max_qty = 25
    quota = quotas.Random(min_qty, max_qty)

    for i in range(0, 1000):
        qty_int = int(quota)
        qty_float = float(quota)

        # Check the quantity falls between the given min/max range
        assert qty_int >= min_qty and qty_int <= max_qty
        assert qty_float >= min_qty and qty_float <= max_qty


import random

__all__ = [
    'Gauss',
    'Random'
    ]


class Quota:
    """
    A base class for implementing variable quota (although the base class
    provides a fixed value and is no different than using an integer or float
    value).

    The Quota class can be safely used as an argument for `Factory`s and
    `Maker`s.
    """

    def __init__(self, quantity):
        self._quantity = quantity

    def __int__(self):
        return int(self._quantity)

    def __float__(self):
        return float(self._quantity)


class Gauss(Quota):
    """
    Return a random quota using a Gaussian distribution.
    """

    def __init__(self, mu, sigma):
        self._mu = mu
        self._sigma = sigma

    def __int__(self):
        return int(float(self))

    def __float__(self):
        return random.gauss(self._mu, self._sigma)


class Random(Quota):
    """
    Return a random quota between two values.
    """

    def __init__(self, min_quantity, max_quantity):
        self._min_quantity = min_quantity
        self._max_quantity = max_quantity

    def __int__(self):
        return random.randint(self._min_quantity, self._max_quantity)

    def __float__(self):
        return random.uniform(self._min_quantity, self._max_quantity)


import datetime
import math
import random
import re

from mongoframes.factory.makers import Maker

__all__ = [
    'DateBetween'
    ]


class DateBetween(Maker):
    """
    Return a date between two dates. Dates can be specified either as
    `datetime.date` instances or as strings of the form:

        "{yesterday|today|tomorrow}{+|-}{no_of_days}"
    """

    def __init__(self, min_date, max_date):
        super().__init__()

        # The date range between which a date will be selected
        self._min_date = min_date
        self._max_date = max_date

    def _assemble(self):
        min_date = self.parse_date(self._min_date)
        max_date = self.parse_date(self._max_date)
        seconds = random.randint(0, int((max_date - min_date).total_seconds()))
        return math.floor(seconds / 86400)

    def _finish(self, value):
        min_date = self.parse_date(self._min_date)
        return min_date + datetime.timedelta(seconds=value * 86400)

    @staticmethod
    def parse_date(d):
        if isinstance(d, datetime.date):
            return d

        # Parse the date string
        result = re.match(
            '(today|tomorrow|yesterday)((\-|\+)(\d+)){0,1}',
            d
            )
        assert result, 'Not a valid date string'

        # Determine the base date
        if result.groups()[0] == 'today':
            d = datetime.date.today()

        elif result.groups()[0] == 'tomorrow':
            d = datetime.date.today() + datetime.timedelta(days=1)

        elif result.groups()[0] == 'yesterday':
            d = datetime.date.today() - datetime.timedelta(days=1)

        # Add any offset
        if result.groups()[1]:
            op = result.groups()[2]
            days = int(result.groups()[3])
            if op == '+':
                d += datetime.timedelta(days=days)
            else:
                d -= datetime.timedelta(days=days)

        return d


from urllib.parse import urlencode

from mongoframes.factory.makers import Maker

__all__ = [
    'ImageURL'
    ]


class ImageURL(Maker):
    """
    Return a fake image URL (by default we use the `fakeimg.pl` service).
    """

    def __init__(self,
        width,
        height,
        background='CCCCCC',
        foreground='8D8D8D',
        options=None,
        service_url='//fakeimg.pl',
        service_formatter=None
        ):
        super().__init__()

        # The size of the image to generate
        self._width = width
        self._height = height

        # The foreground/background colours for the image
        self._background = background
        self._foreground = foreground

        # A dictionary of options used when generating the image
        self._options = options

        # The service URL to use when calling the service
        self._service_url = service_url

        # A formatter function that can produce an image URL for the service
        self._service_formatter = service_formatter or \
                ImageURL._default_service_formatter

    def _assemble(self):
        return self._service_formatter(
            self._service_url,
            int(self._width),
            int(self._height),
            self._background,
            self._foreground,
            self._options
            )

    @staticmethod
    def _default_service_formatter(
        service_url,
        width,
        height,
        background,
        foreground,
        options
        ):
        """Generate an image URL for a service"""

        # Build the base URL
        image_tmp = '{service_url}/{width}x{height}/{background}/{foreground}/'
        image_url = image_tmp.format(
            service_url=service_url,
            width=width,
            height=height,
            background=background,
            foreground=foreground
            )

        # Add any options
        if options:
            image_url += '?' + urlencode(options)

        return image_url
